require("love_reactor/all")
require("speech_bubble")
require("star")

the_player = rxcontainer()
the_player.aggregate("on_destroy")
  .map(function(e)
    the_player.remove(e.member)
  end)

Ship = class({
  sprite = love.graphics.newImage("sprites/ship.psd"),
  shoot = function(self)
    Raytrace.new({
      x = self.x,
      y = 0
    })
    laser_sounds:next():play()
  end,
  explode = function(self)
    local explosion = Explosion.new({
      position = self.position
    })
    self.destroy()
    explosion_sound:play()
    update_stream
      .take_one()
      .delay(2)
      .map(function()
        switcharoo()
      end)
  end
}, {drawable, collidable, movable})

Bullet = class({
  sprite = love.graphics.newImage("sprites/laser.psd"),
  rotation = math.pi,
  depth = -1,
  velocity = {
    x = 0,
    y = 250
  }
}, {drawable, collidable, movable})

Raytrace = class({
  width = 10,
  height = love.window.getHeight() * 2
}, {collidable})

update_stream
  .map(function(e)
    for _, bullet in pairs(Bullet.instances.values()) do
      if bullet.y < 0 or bullet.y > love.window.getHeight() + 100 then
        bullet:destroy()
      end
    end
  end)

death_sounds = sequence({
  love.audio.newSource("audio/die1.ogg"),
  love.audio.newSource("audio/die2.ogg"),
  love.audio.newSource("audio/die3.ogg"),
  love.audio.newSource("audio/die4.ogg"),
  love.audio.newSource("audio/die5.ogg"),
})

Monster = class({
  sprite = love.graphics.newImage("sprites/monster.psd"),
  dying = false,
  cheating = false,
  init = function(self)
    self.sprite = (math.random() < .5 and love.graphics.newImage("sprites/monster.psd")) or love.graphics.newImage("sprites/monster2.psd")
  end,
  hit = function(self)
    if not self.dying and not self.cheating then
      if math.random() < .25 then
        local complaint = SpeechBubble.new({
          x = self.x + 64,
          y = self.y,
          text = table.pick_random({
            "Maaaannnnn",
            "ya got me~",
            "booo",
            "haha aww man",
            "lets get ice cream after"
          })
        })

        update_stream
          .map(function()
            complaint.position = vector.add(self.position, {x=64,y=0})
          end)

        self.on_destroy.map(complaint.destroy)
      end

      if math.random() < .80 then
        tween(self, "position", self.position, {x = -300, y = self.position.y}, 5, linear_interp, function()
          self:destroy()
        end)
      else
        death_sounds:next():play()
        tween(self, "rotation", 0, math.pi / 2, .25)
        local die_distance = 50
        tween(self, "position", self.position, 
                                vector.add(self.position, 
                                  {x = math.frandom(-die_distance, die_distance), y = math.frandom(-die_distance, die_distance)})
                              ,.25)
        self.velocity = {x = 0, y = 0}
      end
      self.dying = true
    end
  end
}, {drawable, movable, collidable})
Monster.speed = 40

update_stream
  .map(function()
    if math.random() < .05 then
      local monster = table.pick_random(alive_monsters.values())
      Bullet.new({
       position = monster.position
      })
    end
  end)

callout_phrases = sequence({
  "hey booo",
  "hit ya!!",
  "you have to die now.",
  "c'mon we hit you your turn is up",
  "cheater!!",
  "stop ittt",
})

oof_sounds = sequence({
  love.audio.newSource("audio/oof1.ogg"),
  love.audio.newSource("audio/oof2.ogg"),
  love.audio.newSource("audio/oof3.ogg"),
  love.audio.newSource("audio/oof4.ogg"),
  love.audio.newSource("audio/oof5.ogg"),
})

local ship_bullet_collision_stream = 
collision_stream
  .map(collision_between(Ship.instances, Bullet.instances))
  .filter(function(e) return e end)

ship_bullet_collision_stream
  .map(function(e)
    local caller = table.pick_random(Monster.instances.values())
    local bubble = SpeechBubble.new({
      position = vector.add(caller.position, {x = 64, y = 0}),
      text = callout_phrases:next()
    })
    update_stream.map(function()
      bubble.position = vector.add(caller.position, {x=64, y = 0})
    end)
    oof_sounds:next():play()
    e.b.destroy()
  end)

honest_monsters = difference(Monster.instances.filter(function(v) return not v.cheating end), the_player)
cheating_monsters = difference(difference(Monster.instances, honest_monsters), the_player)
alive_monsters = difference(Monster.instances.filter(function(v) return not v.dying end), the_player)

collision_stream
  .map(collision_between(cheating_monsters, Raytrace.instances))
  .filter(function(e) return e end)
  .filter(function(e) return not e.a.has_complaint end)
  .map(function(e)
    local complaint = SpeechBubble.new({
      x = e.a.x + 64,
      y = e.a.y,
      text = table.pick_random({
        "you missed!!" ,
        "nuh uh!!",
        "i dodged it!",
        "i put up my deflector shield",
        "wooosh missed me by a mile"
      })
    })
    e.b.destroy()
    e.a.has_complaint = true
    complaint.on_destroy.map(function() e.a.has_complaint = false end)
    update_stream
      .map(function()
        complaint.position = vector.add(e.a.position, {x=64,y=0})
      end)
  end)

collision_stream
  .map(collision_between(Raytrace.instances, intersection(honest_monsters, alive_monsters)))
  .filter(function(e) return e end)
  .buffer(update_stream)
  .filter(function(value) return #value > 0 end)
  .map(function(values)
    if #values > 0 then
      local lowest_monster = table.max(values, function(e) return e.b.y end)
      lowest_monster.b:hit()
      values[1].a.destroy()
    end
  end)

--- MOnSTER MOVEMENT
update_stream
  .map(function()

    local advance = false
    for _, monster in pairs(alive_monsters.values()) do
      if monster.x < 0 and monster.velocity.x < 0 then
        advance = true
        alive_monsters.set("velocity", {x=Monster.speed, y=0})
      end

      if monster.x > love.window.getWidth() and monster.velocity.x > 0 then
        advance = true
        alive_monsters.set("velocity", {x=-Monster.speed, y=0})
      end
    end

    if advance then
      for _, monster in pairs(alive_monsters.values()) do
        tween(monster, "position", monster.position, vector.add(monster.position, {x = 0, y = 32}), 1)
      end
    end
  end)


keyheld_stream
  .buffer(update_stream)
  .map(function(e)
    local keys = table.member_set(e)
    local out_vec = {x = 0, y = 0}
    local mappies = {
      left = {x = -1, y = 0},
      right = {x = 1, y = 0},
      up = {x = 0, y = -1},
      down = {x = 0, y = 1},
    }
    local veccies = table.map(e, function(v) return mappies[v] end)
    local v = table.reduce(veccies, vector.add, {x = 0, y = 0})
    v = vector.normalize(v)
    v = vector.mult(v, 250)

    for _, h in pairs(the_player.values()) do
      h.velocity = v
    end
end)


laser_sounds = sequence({
  love.audio.newSource("audio/laser.ogg"),
  love.audio.newSource("audio/laser2.ogg"),
  love.audio.newSource("audio/laser3.ogg"),
  love.audio.newSource("audio/laser4.ogg"),
  love.audio.newSource("audio/laser5.ogg"),
  love.audio.newSource("audio/laser6.ogg"),
})

keypressed_stream
  .filter(function(e) return e == " " end)
  .map(function()
    for _, ship in pairs(intersection(Ship.instances, the_player).values()) do
      ship:shoot()
    end
  end)

keypressed_stream
  .filter(function(e) return e == " " end)
  .map(function()
    for _, ship in pairs(intersection(Monster.instances, the_player).values()) do
      Bullet.new({
        position = ship.position
      })
    end
  end)

explosion_sound = love.audio.newSource("audio/explosion.ogg")
keypressed_stream
  .filter(function(e) return e == "k" end)
  .filter(function() return not Ship.instances.empty() end)
  .map(function()
    for _, ship in pairs(Ship.instances.values()) do
      ship:explode()
    end
    local monster = table.min(Monster.instances.values(), function(monster) return vector.magnitude(vector.sub(monster.position, {x = love.window.getWidth() / 2, y = love.window.getHeight() / 2 })) end)
    SpeechBubble.new({
      position = monster.position,
      text = "my turn to be the space ship!"
    })
    return "1"
  end)

Explosion = class({
  sprite = love.graphics.newImage("sprites/explosion.psd"),
  width = 64,
  height = 64,
  animations = {
    splode = {
      frames = {1,2,3},
      speed = .1
    }
  },
  start_animation = "splode"
}, {animatable})

Explosion.instances.added
  .delay(1)
  .map(function(value)
    value.destroy()
  end)

function love.load()
  love.graphics.setBackgroundColor(colors.black)
  setup_player_is_ship()

  for i=1,10 do
    Star.new({position = {x = math.random(0, love.window.getWidth()), y = math.random(0, love.window.getHeight())}}, 1)
  end
  for i=1,10 do
    Star.new({position = {x = math.random(0, love.window.getWidth()), y = math.random(0, love.window.getHeight())}}, 2)
  end
  for i=1,10 do
    Star.new({position = {x = math.random(0, love.window.getWidth()), y = math.random(0, love.window.getHeight())}}, 4)
  end
end

function create_enemies()
  local x_nmes = 10
  local y_nmes = 3
  for x=1,x_nmes do
    for y=1,y_nmes do
      local monster = Monster.new({
        x = x * love.window.getWidth() / (x_nmes + 1),
        y = y * 64,
        cheating = (x == 2 and y == 2)
      })
    end
  end
  Monster.instances.set("velocity", {x=-Monster.speed, y=0})
end

switcharoo_called = make_stream()
function switcharoo()
  switcharoo_called.send("called")
  local player_is_ship = intersection(Monster.instances, the_player).empty()
  union(Monster.instances,
        Ship.instances,
        Bullet.instances,
        Raytrace.instances).call('destroy')
  if player_is_ship then
    setup_player_is_badguy()
  else
    setup_player_is_ship()
  end
end

function setup_player_is_ship()
  local ship = Ship.new({
    x = love.window.getWidth() / 2,
    y = love.window.getHeight() - 64
  })

  the_player.add(ship)

  create_enemies()
end

function setup_player_is_badguy()
  create_enemies()
  the_player.add(table.pick_random(Monster.instances.values()))

  local ship = Ship.new({
    x = love.window.getWidth() / 2,
    y = love.window.getHeight() - 64
  })

  -- NPC piloting
  local function move_random()
    local cur_position = ship.position
    local next_position = {x = math.random(0, love.window.getWidth()), y = ship.y}
    local time = vector.magnitude(vector.sub(next_position, cur_position)) / 250
    tween(ship, "position", cur_position, next_position, time, linear_interp, move_random)
  end
  move_random()

  update_stream
    .take_until(switcharoo_called.combine(Ship.instances.aggregate("on_destroy")))
    .map(function()
      if math.random() < 1 / 30 then
        Ship.instances.values()[1]:shoot()
      end
    end)


  ship_bullet_collision_stream
    .take_one()
    .delay(6)
    .map(function()
      for _, ship in pairs(Ship.instances.values()) do
        ship:explode()
      end
    end)
end
